@@ -2,6 +2,7 @@
 #define AMICI_EDATA_H
 
 #include "amici/defines.h"
+#include "amici/vector.h"
 
 #include <vector>
 
@@ -10,20 +11,27 @@ namespace amici {
 class Model;
 class ReturnData;
 
-/** @brief ExpData carries all information about experimental or
- * condition-specific data */
+/**
+ * @brief ExpData carries all information about experimental or
+ * condition-specific data
+ */
 class ExpData {
 
   public:
-    /** @brief default constructor */
-    ExpData();
+    /**
+     * @brief default constructor
+     */
+    ExpData() = default;
 
-    /** @brief Copy constructor, needs to be declared to be generated in
-     * swig*/
+    /**
+     * @brief Copy constructor, needs to be declared to be generated in
+     * swig
+     */
     ExpData(const ExpData &) = default;
 
     /**
      * @brief constructor that only initializes dimensions
+     *
      * @param nytrue
      * @param nztrue
      * @param nmaxevent
@@ -79,8 +87,8 @@ class ExpData {
     explicit ExpData(const Model &model);
 
     /**
-     * @brief constructor that initializes with returnData, adds noise
-     * according to specified sigmas
+     * @brief constructor that initializes with returnData, adds noise according
+     * to specified sigmas
      *
      * @param rdata return data pointer with stored simulation results
      * @param sigma_y scalar standard deviations for all observables
@@ -89,12 +97,12 @@ class ExpData {
     ExpData(const ReturnData &rdata, realtype sigma_y, realtype sigma_z);
 
     /**
-     * @brief constructor that initializes with returnData, adds noise
-     * according to specified sigmas
+     * @brief constructor that initializes with returnData, adds noise according
+     * to specified sigmas
      *
      * @param rdata return data pointer with stored simulation results
-     * @param sigma_y vector of standard deviations for observables (dimension:
-     * nytrue or nt x nytrue, row-major)
+     * @param sigma_y vector of standard deviations for observables
+     * (dimension: nytrue or nt x nytrue, row-major)
      * @param sigma_z vector of standard deviations for event observables
      * (dimension: nztrue or nmaxevent x nztrue, row-major)
      */
@@ -149,6 +157,7 @@ class ExpData {
      * @brief get function that returns timepoint at index
      *
      * @param it timepoint index
+     *
      * @return timepoint timepoint at index
      */
     realtype getTimepoint(int it) const;
@@ -174,6 +183,7 @@ class ExpData {
      *
      * @param it time index
      * @param iy observable index
+     *
      * @return boolean specifying if data was set
      */
     bool isSetObservedData(int it, int iy) const;
@@ -189,6 +199,7 @@ class ExpData {
      * @brief get function that returns a pointer to observed data at index
      *
      * @param it timepoint index
+     *
      * @return pointer to observed data at index (dimension: nytrue)
      */
     const realtype *getObservedDataPtr(int it) const;
@@ -208,7 +219,7 @@ class ExpData {
      *
      * @param stdDev standard deviation (dimension: scalar)
      */
-    void setObservedDataStdDev(const realtype stdDev);
+    void setObservedDataStdDev(realtype stdDev);
 
     /**
      * @brief set function that copies standard deviation of observed data for
@@ -228,7 +239,7 @@ class ExpData {
      * @param stdDev standard deviation (dimension: scalar)
      * @param iy observed data index
      */
-    void setObservedDataStdDev(const realtype stdDev, int iy);
+    void setObservedDataStdDev(realtype stdDev, int iy);
 
     /**
      * @brief get function that checks whether standard deviation of data at
@@ -297,6 +308,7 @@ class ExpData {
      * occurence
      *
      * @param ie event occurence
+     *
      * @return pointer to observed event data at ieth occurence
      */
     const realtype *getObservedEventsPtr(int ie) const;
@@ -316,7 +328,7 @@ class ExpData {
      *
      * @param stdDev standard deviation (dimension: scalar)
      */
-    void setObservedEventsStdDev(const realtype stdDev);
+    void setObservedEventsStdDev(realtype stdDev);
 
     /**
      * @brief set function that copies standard deviation of observed data for
@@ -337,7 +349,7 @@ class ExpData {
      * @param stdDev standard deviation (dimension: scalar)
      * @param iz observed data index
      */
-    void setObservedEventsStdDev(const realtype stdDev, int iz);
+    void setObservedEventsStdDev(realtype stdDev, int iz);
 
     /**
      * @brief get function that checks whether standard deviation of even data
@@ -362,6 +374,7 @@ class ExpData {
      * observed event data at ieth occurence
      *
      * @param ie event occurence
+     *
      * @return pointer to standard deviation of observed event data at ieth
      * occurence
      */
@@ -371,32 +384,41 @@ class ExpData {
      * @brief condition-specific fixed parameters of size Model::nk() or empty
      */
     std::vector<realtype> fixedParameters;
-    /** @brief condition-specific fixed parameters for pre-equilibration of size
+    /**
+     * @brief condition-specific fixed parameters for pre-equilibration of size
      * Model::nk() or empty. Overrides Solver::newton_preeq
      */
     std::vector<realtype> fixedParametersPreequilibration;
-    /** @brief condition-specific fixed parameters for pre-simulation of
+    /**
+     * @brief condition-specific fixed parameters for pre-simulation of
      * size Model::nk() or empty.
      */
     std::vector<realtype> fixedParametersPresimulation;
 
-    /** @brief condition-specific parameters of size Model::np() or empty */
+    /**
+     * @brief condition-specific parameters of size Model::np() or empty
+     */
     std::vector<realtype> parameters;
-    /** @brief condition-specific initial conditions of size Model::nx() or
+    /**
+     * @brief condition-specific initial conditions of size Model::nx() or
      * empty
      */
     std::vector<realtype> x0;
-    /** @brief condition-specific initial condition sensitivities of size
+    /**
+     * @brief condition-specific initial condition sensitivities of size
      * Model::nx() * Model::nplist(), Model::nx() * ExpDataplist.size(), if
      * ExpData::plist is not empty, or empty
      */
     std::vector<realtype> sx0;
-    /** @brief condition-specific parameter scales of size Model::np()
+    /**
+     * @brief condition-specific parameter scales of size Model::np()
      */
     std::vector<ParameterScaling> pscale;
-    /** @brief condition-specific parameter list */
+    /**
+     * @brief condition-specific parameter list
+     */
     std::vector<int> plist;
-    
+
     /**
      * @brief duration of pre-simulation
      * if this is > 0, presimualation will be performed from
@@ -405,6 +427,11 @@ class ExpData {
      */
     realtype t_presim = 0;
 
+    /** flag indicating whether reinitialization of states depending on
+     *  fixed parameters is activated
+     */
+    bool reinitializeFixedParameterInitialStates = false;
+
   protected:
     /**
      * @brief resizes observedData, observedDataStdDev, observedEvents and
@@ -442,26 +469,31 @@ class ExpData {
                               const char *fieldname) const;
 
     /** @brief number of observables */
-    int nytrue_;
+    int nytrue_{0};
 
     /** @brief number of event observables */
-    int nztrue_;
+    int nztrue_{0};
 
     /** @brief maximal number of event occurences */
-    int nmaxevent_;
+    int nmaxevent_{0};
 
     /** @brief observation timepoints (dimension: nt) */
     std::vector<realtype> ts;
 
     /** @brief observed data (dimension: nt x nytrue, row-major) */
     std::vector<realtype> observedData;
-    /** @brief standard deviation of observed data (dimension: nt x nytrue,
-     * row-major) */
+    /**
+     * @brief standard deviation of observed data (dimension: nt x nytrue,
+     * row-major)
+     */
     std::vector<realtype> observedDataStdDev;
 
-    /** @brief observed events (dimension: nmaxevents x nztrue, row-major) */
+    /**
+     * @brief observed events (dimension: nmaxevents x nztrue, row-major)
+     */
     std::vector<realtype> observedEvents;
-    /** @brief standard deviation of observed events/roots
+    /**
+     * @brief standard deviation of observed events/roots
      * (dimension: nmaxevents x nztrue, row-major)
      */
     std::vector<realtype> observedEventsStdDev;
@@ -482,7 +514,7 @@ void checkSigmaPositivity(std::vector<realtype> const &sigmaVector,
  * @param sigma input to be checked
  * @param sigmaName name of the input
  */
-void checkSigmaPositivity(const realtype sigma, const char *sigmaName);
+void checkSigmaPositivity(realtype sigma, const char *sigmaName);
 
 /**
  * @brief The ConditionContext class applies condition-specific amici::Model
@@ -497,7 +529,7 @@ class ConditionContext {
      * @param model
      * @param edata
      */
-    ConditionContext(Model *model, const ExpData *edata = nullptr);
+    explicit ConditionContext(Model *model, const ExpData *edata = nullptr);
 
     ~ConditionContext();
 
@@ -526,7 +558,7 @@ class ConditionContext {
     std::vector<realtype> originalTimepoints;
     std::vector<int> originalParameterList;
     std::vector<amici::ParameterScaling> originalScaling;
-    
+    bool originalReinitializeFixedParameterInitialStates;
 };
 
 } // namespace amici
