@@ -97,33 +97,30 @@ class SteadystateProblem {
      * @param model pointer to the AMICI model object
      * @return solver instance
      */
-    std::unique_ptr<Solver> createSteadystateSimSolver(Solver *solver,
-                                                       Model *model);
+    std::unique_ptr<Solver> createSteadystateSimSolver(const Solver *solver,
+                                                       Model *model) const;
 
-    /** default constructor
-     * @param t pointer to time variable
-     * @param x pointer to state variables
-     * @param sx pointer to state sensitivity variables
+    /**
+     * @brief constructor
+     * @param solver pointer to Solver instance
+     * @param x0 initial state
      */
-    SteadystateProblem(realtype *t, AmiVector *x, AmiVectorArray *sx) :
-    delta(x->getLength()),
-    ewt(x->getLength()),
-    rel_x_newton(x->getLength()),
-    x_newton(x->getLength()),
-    x_old(x->getLength()),
-    dx(x->getLength()),
-    xdot(x->getLength()),
-    xdot_old(x->getLength()),
-    sdx(x->getLength(),sx->getLength())
-    {
-        this->t = t;
-        this->x = x;
-        this->sx = sx;
-    }
+    explicit SteadystateProblem(const Solver *solver, const AmiVector &x0);
+    
+    /**
+     * @brief routine that writes solutions of steadystate problem to target
+     vectors
+     * @param t final timepoint
+     * @param x steadystate state
+     * @param sx steadystate state sensitivity
+     */
+    void writeSolution(realtype *t, AmiVector &x, AmiVectorArray &sx) const;
+    
 
   private:
-    realtype *t;
-    /** newton step? */
+    /** time variable for simulation steadystate finding */
+    realtype t;
+    /** newton step */
     AmiVector delta;
     /** error weights */
     AmiVector ewt;
@@ -132,7 +129,7 @@ class SteadystateProblem {
     /** container for absolute error calcuation? */
     AmiVector x_newton;
     /** state vector */
-    AmiVector *x;
+    AmiVector x;
     /** old state vector */
     AmiVector x_old;
     /** differential state vector */
@@ -142,7 +139,7 @@ class SteadystateProblem {
     /** old time derivative state vector */
     AmiVector xdot_old;
     /** state sensitivities */
-    AmiVectorArray *sx;
+    AmiVectorArray sx;
     /** state differential sensitivities */
     AmiVectorArray sdx;
 
