@@ -42,18 +42,24 @@ extern void dJydsigmay_TPL_MODELNAME(realtype *dJydsigmay, const int iy,
                                      const realtype *p, const realtype *k,
                                      const realtype *y, const realtype *sigmay,
                                      const realtype *my);
-extern void dJydy_TPL_MODELNAME(realtype *dJydy, const int iy,
-                                const realtype *p, const realtype *k,
-                                const realtype *y, const realtype *sigmay,
-                                const realtype *my);
+TPL_DJYDY_DEF
+TPL_DJYDY_COLPTRS_DEF
+TPL_DJYDY_ROWVALS_DEF
 TPL_DWDP_DEF
+TPL_DWDP_COLPTRS_DEF
+TPL_DWDP_ROWVALS_DEF
 TPL_DWDX_DEF
 TPL_DWDX_COLPTRS_DEF
 TPL_DWDX_ROWVALS_DEF
 TPL_DXDOTDW_DEF
 TPL_DXDOTDW_COLPTRS_DEF
 TPL_DXDOTDW_ROWVALS_DEF
-TPL_DXDOTDP_DEF
+TPL_DXDOTDP_EXPLICIT_DEF
+TPL_DXDOTDP_EXPLICIT_COLPTRS_DEF
+TPL_DXDOTDP_EXPLICIT_ROWVALS_DEF
+TPL_DXDOTDP_IMPLICIT_COLPTRS_DEF
+TPL_DXDOTDP_IMPLICIT_ROWVALS_DEF
+
 extern void dydx_TPL_MODELNAME(realtype *dydx, const realtype t,
                                const realtype *x, const realtype *p,
                                const realtype *k, const realtype *h,
@@ -116,6 +122,7 @@ class Model_TPL_MODELNAME : public amici::Model_ODE {
               TPL_NDWDX,                                   // ndwdx
               TPL_NDWDP,                                   // ndwdp
               TPL_NDXDOTDW,                                // ndxdotdw
+              TPL_NDJYDY,                                  // ndjydy
               TPL_NNZ,                                     // nnz
               TPL_UBW,                                     // ubw
               TPL_LBW,                                     // lbw
@@ -124,7 +131,10 @@ class Model_TPL_MODELNAME : public amici::Model_ODE {
               std::vector<realtype>{TPL_FIXED_PARAMETERS}, // fixedParameters
               std::vector<int>{},                          // plist
               std::vector<realtype>(TPL_NX_SOLVER, 0.0),   // idlist
-              std::vector<int>{}                           // z2event
+              std::vector<int>{},                          // z2event
+              true,                                        // pythonGenerated
+              TPL_NDXDOTDP_EXPLICIT,                       // ndxdotdp_explicit
+              TPL_NDXDOTDP_IMPLICIT                        // ndxdotdp_implicit
           ) {}
 
     /**
@@ -186,15 +196,15 @@ class Model_TPL_MODELNAME : public amici::Model_ODE {
     }
 
     TPL_JSPARSE_IMPL
-    
+
     TPL_JSPARSE_COLPTRS_IMPL
-    
+
     TPL_JSPARSE_ROWVALS_IMPL
 
     TPL_JSPARSEB_IMPL
-    
+
     TPL_JSPARSEB_COLPTRS_IMPL
-    
+
     TPL_JSPARSEB_ROWVALS_IMPL
 
     /** model specific implementation of fJrz
@@ -280,21 +290,6 @@ class Model_TPL_MODELNAME : public amici::Model_ODE {
         dJydsigmay_TPL_MODELNAME(dJydsigma, iy, p, k, y, sigmay, my);
     }
 
-    /** model specific implementation of fdJydy
-     * @param dJydy partial derivative of time-resolved measurement negative
-     *log-likelihood Jy
-     * @param iy output index
-     * @param p parameter vector
-     * @param k constant vector
-     * @param y model output at timepoint
-     * @param sigmay measurement standard deviation at timepoint
-     * @param my measurement at timepoint
-     **/
-    virtual void fdJydy(realtype *dJydy, const int iy, const realtype *p,
-                        const realtype *k, const realtype *y,
-                        const realtype *sigmay, const realtype *my) override {
-        dJydy_TPL_MODELNAME(dJydy, iy, p, k, y, sigmay, my);
-    }
 
     /** model specific implementation of fdJzdsigma
      * @param dJzdsigma Sensitivity of event measurement
@@ -454,18 +449,36 @@ class Model_TPL_MODELNAME : public amici::Model_ODE {
                             const realtype *p, const realtype *k,
                             const int ip) override {}
 
+    TPL_DJYDY_IMPL
+    TPL_DJYDY_COLPTRS_IMPL
+    TPL_DJYDY_ROWVALS_IMPL
+
     TPL_DWDP_IMPL
     
+    TPL_DWDP_COLPTRS_IMPL
+    
+    TPL_DWDP_ROWVALS_IMPL
+
     TPL_DWDX_IMPL
+
+    TPL_DWDX_COLPTRS_IMPL
+    
+    TPL_DWDX_ROWVALS_IMPL
     
     TPL_DXDOTDW_IMPL
-    
     TPL_DXDOTDW_COLPTRS_IMPL
-    
     TPL_DXDOTDW_ROWVALS_IMPL
-    
-    TPL_DXDOTDP_IMPL
 
+    TPL_DXDOTDP_EXPLICIT_IMPL
+    
+    TPL_DXDOTDP_EXPLICIT_COLPTRS_IMPL
+    
+    TPL_DXDOTDP_EXPLICIT_ROWVALS_IMPL
+
+    TPL_DXDOTDP_IMPLICIT_COLPTRS_IMPL
+    
+    TPL_DXDOTDP_IMPLICIT_ROWVALS_IMPL
+    
     /** model specific implementation of fdydx
      * @param dydx partial derivative of observables y w.r.t. model states x
      * @param t current time
@@ -716,10 +729,14 @@ class Model_TPL_MODELNAME : public amici::Model_ODE {
                     const realtype *h) override {}
 
     TPL_X_RDATA_IMPL
-    
+
     TPL_X_SOLVER_IMPL
-    
+
     TPL_TOTAL_CL_IMPL
+    
+    std::string getName() const override {
+        return "TPL_MODELNAME";
+    }
 
     /**
      * @brief Get names of the model parameters
@@ -812,10 +829,6 @@ class Model_TPL_MODELNAME : public amici::Model_ODE {
     virtual const std::string getAmiciCommit() const override {
         return "TPL_AMICI_COMMIT_STRING";
     }
-    
-    virtual bool wasPythonGenerated() const override {
-        return true;
-    }
 };
 
 #endif /* _amici_TPL_MODELNAME_h */
