@@ -1,9 +1,7 @@
 #include "amici/newton_solver.h"
 
-#include "amici/defines.h"
 #include "amici/model.h"
 #include "amici/solver.h"
-#include "amici/vector.h"
 #include "amici/steadystateproblem.h"
 #include "amici/forwardproblem.h"
 #include "amici/rdata.h"
@@ -18,17 +16,10 @@
 
 namespace amici {
 
-NewtonSolver::NewtonSolver(realtype *t, AmiVector *x, Model *model, ReturnData *rdata)
-    : model(model), rdata(rdata), xdot(x->getLength()), dx(x->getLength())
+NewtonSolver::NewtonSolver(realtype *t, AmiVector *x, Model *model,
+                           ReturnData *rdata)
+    : model(model), rdata(rdata), xdot(model->nx_solver), dx(model->nx_solver)
     {
-    /**
-     * default constructor, initializes all members with the provided objects
-     *
-     * @param t pointer to time variable
-     * @param x pointer to state variables
-     * @param model pointer to the AMICI model object
-     * @param rdata pointer to the return data object
-     */
     this->t = t;
     this->x = x;
 }
@@ -37,24 +28,8 @@ NewtonSolver::NewtonSolver(realtype *t, AmiVector *x, Model *model, ReturnData *
 
 std::unique_ptr<NewtonSolver> NewtonSolver::getSolver(
         realtype *t, AmiVector *x, LinearSolver linsolType, Model *model,
-        ReturnData *rdata, int maxlinsteps, int maxsteps, double atol, double rtol) {
-    /**
-     * Tries to determine the steady state of the ODE system by a Newton
-     * solver, uses forward intergration, if the Newton solver fails,
-     * restarts Newton solver, if integration fails.
-     * Computes steady state sensitivities
-     *
-     * @param t pointer to time variable
-     * @param x pointer to state variables
-     * @param linsolType integer indicating which linear solver to use
-     * @param model pointer to the AMICI model object
-     * @param rdata pointer to the return data object
-     * @param maxlinsteps maximum number of allowed linear steps per Newton step for steady state computation
-     * @param maxsteps maximum number of allowed Newton steps for steady state computation
-     * @param atol absolute tolerance
-     * @param rtol relative tolerance
-     * @return solver NewtonSolver according to the specified linsolType
-     */
+        ReturnData *rdata, int maxlinsteps, int maxsteps, double atol, double rtol,
+        NewtonDampingFactorMode dampingFactorMode, double dampingFactorLowerBound) {
 
     std::unique_ptr<NewtonSolver> solver;
 
@@ -89,6 +64,8 @@ std::unique_ptr<NewtonSolver> NewtonSolver::getSolver(
         throw NewtonFailure(AMICI_NOT_IMPLEMENTED, "getSolver");
 
     /* SPARSE SOLVERS */
+    case LinearSolver::SuperLUMT:
+        throw NewtonFailure(AMICI_NOT_IMPLEMENTED, "getSolver");
     case LinearSolver::KLU:
         solver.reset(new NewtonSolverSparse(t, x, model, rdata));
         break;
@@ -100,46 +77,60 @@ std::unique_ptr<NewtonSolver> NewtonSolver::getSolver(
     solver->rtol = rtol;
     solver->maxlinsteps = maxlinsteps;
     solver->maxsteps = maxsteps;
+    solver->dampingFactorMode = dampingFactorMode;
+    solver->dampingFactorLowerBound = dampingFactorLowerBound;
 
     return solver;
 }
 
 /* ------------------------------------------------------------------------- */
 
-void NewtonSolver::getStep(int ntry, int nnewt, AmiVector *delta) {
-    /**
-     * Computes the solution of one Newton iteration
-     *
-     * @param ntry integer newton_try integer start number of Newton solver
-     * (1 or 2)
-     * @param nnewt integer number of current Newton step
-     * @param delta containing the RHS of the linear system, will be
-     * overwritten by solution to the linear system
-     */
-
+void NewtonSolver::getStep(int ntry, int nnewt, AmiVector &delta) {
     this->prepareLinearSystem(ntry, nnewt);
 
-    delta->minus();
+    delta.minus();
     this->solveLinearSystem(delta);
 }
 
 /* ------------------------------------------------------------------------- */
 
-void NewtonSolver::computeNewtonSensis(AmiVectorArray *sx) {
-    /**
-     * Computes steady state sensitivities
-     *
-     * @param sx pointer to state variable sensitivities
-     */
+void NewtonSolver::computeNewtonSensis(AmiVectorArray &sx) {
     prepareLinearSystem(0, -1);
+    model->fdxdotdp(*t, *x, dx);
 
-    model->fdxdotdp(*t, x, &dx);
-    for (int ip = 0; ip < model->nplist(); ip++) {
+    if (model->pythonGenerated) {
+        for (int ip = 0; ip < model->nplist(); ip++) {
+            N_VConst(0.0, sx.getNVector(ip));
 
-        for (int ix = 0; ix < model->nx_solver; ix++) {
-            sx->at(ix,ip) = -model->dxdotdp.at(ix, ip);
+            // copy explicit version
+            if (model->ndxdotdp_explicit > 0) {
+                auto col = model->dxdotdp_explicit.indexptrs();
+                auto row = model->dxdotdp_explicit.indexvals();
+                auto data_ptr = model->dxdotdp_explicit.data();
+                for (sunindextype iCol = col[model->plist(ip)];
+                     iCol < col[model->plist(ip) + 1]; ++iCol)
+                    sx.at(row[iCol], ip) -= data_ptr[iCol];
+            }
+
+            // copy implicit version
+            if (model->ndxdotdp_implicit > 0) {
+                auto col = model->dxdotdp_implicit.indexptrs();
+                auto row = model->dxdotdp_implicit.indexvals();
+                auto data_ptr = model->dxdotdp_implicit.data();
+                for (sunindextype iCol = col[model->plist(ip)];
+                     iCol < col[model->plist(ip) + 1]; ++iCol)
+                    sx.at(row[iCol], ip) -= data_ptr[iCol];
+            }
+
+            solveLinearSystem(sx[ip]);
+        }
+    } else {
+        for (int ip = 0; ip < model->nplist(); ip++) {
+            for (int ix = 0; ix < model->nx_solver; ix++)
+                sx.at(ix,ip) = -model->dxdotdp.at(ix, ip);
+
+            solveLinearSystem(sx[ip]);
         }
-        solveLinearSystem(&((*sx)[ip]));
     }
 }
 /* ------------------------------------------------------------------------- */
@@ -147,21 +138,12 @@ void NewtonSolver::computeNewtonSensis(AmiVectorArray *sx) {
 /* ------------------------------------------------------------------------- */
 
 /* Derived class for dense linear solver */
-NewtonSolverDense::NewtonSolverDense(realtype *t, AmiVector *x, Model *model, ReturnData *rdata)
+NewtonSolverDense::NewtonSolverDense(realtype *t, AmiVector *x, Model *model,
+                                     ReturnData *rdata)
     : NewtonSolver(t, x, model, rdata),
       Jtmp(model->nx_solver, model->nx_solver),
       linsol(SUNLinSol_Dense(x->getNVector(), Jtmp.get()))
 {
-    /**
-     * default constructor, initializes all members with the provided objects
-     * and
-     * initializes temporary storage objects
-     *
-     * @param t pointer to time variable
-     * @param x pointer to state variables
-     * @param model pointer to the AMICI model object
-     * @param rdata pointer to the return data object
-     */
     int status = SUNLinSolInitialize_Dense(linsol);
     if(status != AMICI_SUCCESS)
         throw NewtonFailure(status, "SUNLinSolInitialize_Dense");
@@ -170,16 +152,7 @@ NewtonSolverDense::NewtonSolverDense(realtype *t, AmiVector *x, Model *model, Re
 /* ------------------------------------------------------------------------- */
 
 void NewtonSolverDense::prepareLinearSystem(int  /*ntry*/, int  /*nnewt*/) {
-    /**
-     * Writes the Jacobian for the Newton iteration and passes it to the linear
-     * solver
-     *
-     * @param ntry integer newton_try integer start number of Newton solver
-     * (1 or 2)
-     * @param nnewt integer number of current Newton step
-     */
-
-    model->fJ(*t, 0.0, x, &dx, &xdot, Jtmp.get());
+    model->fJ(*t, 0.0, *x, dx, xdot, Jtmp.get());
     int status = SUNLinSolSetup_Dense(linsol, Jtmp.get());
     if(status != AMICI_SUCCESS)
         throw NewtonFailure(status, "SUNLinSolSetup_Dense");
@@ -187,19 +160,12 @@ void NewtonSolverDense::prepareLinearSystem(int  /*ntry*/, int  /*nnewt*/) {
 
 /* ------------------------------------------------------------------------- */
 
-void NewtonSolverDense::solveLinearSystem(AmiVector *rhs) {
-    /**
-     * Solves the linear system for the Newton step
-     *
-     * @param rhs containing the RHS of the linear system, will be
-     * overwritten by solution to the linear system
-     */
-
+void NewtonSolverDense::solveLinearSystem(AmiVector &rhs) {
     int status = SUNLinSolSolve_Dense(linsol, Jtmp.get(),
-                                      rhs->getNVector(), rhs->getNVector(),
+                                      rhs.getNVector(), rhs.getNVector(),
                                       0.0);
     // last argument is tolerance and does not have any influence on result
-    
+
     if(status != AMICI_SUCCESS)
         throw NewtonFailure(status, "SUNLinSolSolve_Dense");
 }
@@ -216,20 +182,12 @@ NewtonSolverDense::~NewtonSolverDense() {
 /* ------------------------------------------------------------------------- */
 
 /* Derived class for sparse linear solver */
-NewtonSolverSparse::NewtonSolverSparse(realtype *t, AmiVector *x, Model *model, ReturnData *rdata)
+NewtonSolverSparse::NewtonSolverSparse(realtype *t, AmiVector *x, Model *model,
+                                       ReturnData *rdata)
     : NewtonSolver(t, x, model, rdata),
       Jtmp(model->nx_solver, model->nx_solver, model->nnz, CSC_MAT),
       linsol(SUNKLU(x->getNVector(), Jtmp.get()))
 {
-    /**
-     * default constructor, initializes all members with the provided objects,
-     * initializes temporary storage objects and the klu solver
-     *
-     * @param t pointer to time variable
-     * @param x pointer to state variables
-     * @param model pointer to the AMICI model object
-     * @param rdata pointer to the return data object
-     */
     int status = SUNLinSolInitialize_KLU(linsol);
     if(status != AMICI_SUCCESS)
         throw NewtonFailure(status, "SUNLinSolInitialize_KLU");
@@ -238,38 +196,21 @@ NewtonSolverSparse::NewtonSolverSparse(realtype *t, AmiVector *x, Model *model,
 /* ------------------------------------------------------------------------- */
 
 void NewtonSolverSparse::prepareLinearSystem(int  /*ntry*/, int  /*nnewt*/) {
-    /**
-     * Writes the Jacobian for the Newton iteration and passes it to the linear
-     * solver
-     *
-     * @param ntry integer newton_try integer start number of Newton solver
-     * (1 or 2)
-     * @param nnewt integer number of current Newton step
-     */
-
     /* Get sparse Jacobian */
-    model->fJSparse(*t, 0.0, x, &dx, &xdot, Jtmp.get());
+    model->fJSparse(*t, 0.0, *x, dx, xdot, Jtmp.get());
     int status = SUNLinSolSetup_KLU(linsol, Jtmp.get());
     if(status != AMICI_SUCCESS)
         throw NewtonFailure(status, "SUNLinSolSetup_KLU");
-} // namespace amici
+}
 
 /* ------------------------------------------------------------------------- */
 
-void NewtonSolverSparse::solveLinearSystem(AmiVector *rhs) {
-    /**
-     * Solves the linear system for the Newton step
-     *
-     * @param rhs containing the RHS of the linear system,will be
-     * overwritten by solution to the linear system
-     */
-
+void NewtonSolverSparse::solveLinearSystem(AmiVector &rhs) {
     /* Pass pointer to the linear solver */
     int status = SUNLinSolSolve_KLU(linsol, Jtmp.get(),
-                                    rhs->getNVector(), rhs->getNVector(),
-                                    0.0);
+                                    rhs.getNVector(), rhs.getNVector(), 0.0);
     // last argument is tolerance and does not have any influence on result
-    
+
     if(status != AMICI_SUCCESS)
         throw NewtonFailure(status, "SUNLinSolSolve_Dense");
 }
@@ -285,78 +226,42 @@ NewtonSolverSparse::~NewtonSolverSparse() {
 /* - Iterative linear solver------------------------------------------------ */
 /* ------------------------------------------------------------------------- */
 
-NewtonSolverIterative::NewtonSolverIterative(realtype *t, AmiVector *x, Model *model, ReturnData *rdata)
-    : NewtonSolver(t, x, model, rdata), ns_p(model->nx_solver), ns_h(model->nx_solver),
-    ns_t(model->nx_solver), ns_s(model->nx_solver), ns_r(model->nx_solver), ns_rt(model->nx_solver), ns_v(model->nx_solver),
-    ns_Jv(model->nx_solver), ns_tmp(model->nx_solver), ns_Jdiag(model->nx_solver)
+NewtonSolverIterative::NewtonSolverIterative(realtype *t, AmiVector *x,
+                                             Model *model, ReturnData *rdata)
+    : NewtonSolver(t, x, model, rdata), ns_p(model->nx_solver),
+    ns_h(model->nx_solver), ns_t(model->nx_solver), ns_s(model->nx_solver),
+    ns_r(model->nx_solver), ns_rt(model->nx_solver), ns_v(model->nx_solver),
+    ns_Jv(model->nx_solver), ns_tmp(model->nx_solver),
+    ns_Jdiag(model->nx_solver)
     {
-    /**
-     * default constructor, initializes all members with the provided objects
-     * @param t pointer to time variable
-     * @param x pointer to state variables
-     * @param model pointer to the AMICI model object
-     * @param rdata pointer to the return data object
-     */
 }
 
 /* ------------------------------------------------------------------------- */
 
 void NewtonSolverIterative::prepareLinearSystem(int ntry, int nnewt) {
-    /**
-     * Writes the Jacobian for the Newton iteration and passes it to the linear
-     * solver.
-     * Also wraps around getSensis for iterative linear solver.
-     *
-     * @param ntry integer newton_try integer start number of Newton solver
-     * (1 or 2)
-     * @param nnewt integer number of current Newton step
-     */
-
     newton_try = ntry;
     i_newton = nnewt;
     if (nnewt == -1) {
-        throw AmiException("Linear solver SPBCG does not support sensitivity computation for steady state problems.");
+        throw AmiException("Linear solver SPBCG does not support sensitivity "
+                           "computation for steady state problems.");
     }
 }
 
 /* ------------------------------------------------------------------------- */
 
-void NewtonSolverIterative::solveLinearSystem(AmiVector *rhs) {
-    /**
-     * Solves the linear system for the Newton step by passing it to
-     * linsolveSPBCG
-     *
-     * @param rhs containing the RHS of the linear system, will be
-     * overwritten by solution to the linear system
-     */
-
+void NewtonSolverIterative::solveLinearSystem(AmiVector &rhs) {
     linsolveSPBCG(newton_try, i_newton, rhs);
-    rhs->minus();
+    rhs.minus();
 }
 
 
-void NewtonSolverIterative::linsolveSPBCG(int ntry,int nnewt, AmiVector *ns_delta) {
-    /**
-     * Iterative linear solver created from SPILS BiCG-Stab.
-     * Solves the linear system within each Newton step if iterative solver is
-     * chosen.
-     *
-     * @param ntry integer newton_try integer start number of Newton solver
-     * (1 or 2)
-     * @param nnewt integer number of current Newton step
-     * @param ns_delta Newton step
-     */
-
-    double rho;
-    double alpha;
-    double omega;
-    double res;
-
-    xdot = *ns_delta;
+void NewtonSolverIterative::linsolveSPBCG(int ntry, int nnewt,
+                                          AmiVector &ns_delta) {
+    xdot = ns_delta;
     xdot.minus();
 
     // Get the diagonal of the Jacobian for preconditioning
-    model->fJDiag(*t, &ns_Jdiag, 0.0, x, &dx);
+    model->fJDiag(*t, ns_Jdiag, 0.0, *x, dx);
 
     // Ensure positivity of entries in ns_Jdiag
     ns_p.set(1.0);
@@ -368,19 +273,18 @@ void NewtonSolverIterative::linsolveSPBCG(int ntry,int nnewt, AmiVector *ns_delt
     // Initialize for linear solve
     ns_p.reset();
     ns_v.reset();
-    ns_delta->reset();
+    ns_delta.reset();
     ns_tmp.reset();
-    rho = 1.0;
-    omega = 1.0;
-    alpha = 1.0;
+    double rho = 1.0;
+    double omega = 1.0;
+    double alpha = 1.0;
 
     // can be set to 0 at the moment
-    model->fJv(*t, x, &dx, &xdot, ns_delta, &ns_Jv, 0.0);
+    model->fJv(*t, *x, dx, xdot, ns_delta, ns_Jv, 0.0);
 
     // ns_r = xdot - ns_Jv;
     N_VLinearSum(-1.0, ns_Jv.getNVector(), 1.0, xdot.getNVector(), ns_r.getNVector());
     N_VDiv(ns_r.getNVector(), ns_Jdiag.getNVector(), ns_r.getNVector());
-    res = sqrt(N_VDotProd(ns_r.getNVector(), ns_r.getNVector()));
     ns_rt = ns_r;
 
     for (int i_linstep = 0; i_linstep < maxlinsteps;
@@ -395,32 +299,34 @@ void NewtonSolverIterative::linsolveSPBCG(int ntry,int nnewt, AmiVector *ns_delt
         N_VLinearSum(1.0, ns_r.getNVector(), beta, ns_p.getNVector(), ns_p.getNVector());
 
         // ns_v = J * ns_p
-        model->fJv(*t, x, &dx, &xdot, &ns_p, &ns_v, 0.0);
+        model->fJv(*t, *x, dx, xdot, ns_p, ns_v, 0.0);
         N_VDiv(ns_v.getNVector(), ns_Jdiag.getNVector(), ns_v.getNVector());
 
         // Compute factor
         alpha = rho / N_VDotProd(ns_rt.getNVector(), ns_v.getNVector());
 
         // ns_h = ns_delta + alpha * ns_p;
-        N_VLinearSum(1.0, ns_delta->getNVector(), alpha, ns_p.getNVector(), ns_h.getNVector());
+        N_VLinearSum(1.0, ns_delta.getNVector(), alpha, ns_p.getNVector(),
+                     ns_h.getNVector());
         // ns_s = ns_r - alpha * ns_v;
         N_VLinearSum(1.0, ns_r.getNVector(), -alpha, ns_v.getNVector(), ns_s.getNVector());
 
         // ns_t = J * ns_s
-        model->fJv(*t, x, &dx, &xdot, &ns_s, &ns_t, 0.0);
+        model->fJv(*t, *x, dx, xdot, ns_s, ns_t, 0.0);
         N_VDiv(ns_t.getNVector(), ns_Jdiag.getNVector(), ns_t.getNVector());
 
         // Compute factor
         omega = N_VDotProd(ns_t.getNVector(), ns_s.getNVector()) / N_VDotProd(ns_t.getNVector(), ns_t.getNVector());
 
         // ns_delta = ns_h + omega * ns_s;
-        N_VLinearSum(1.0, ns_h.getNVector(), omega, ns_s.getNVector(), ns_delta->getNVector());
+        N_VLinearSum(1.0, ns_h.getNVector(), omega, ns_s.getNVector(),
+                     ns_delta.getNVector());
         // ns_r = ns_s - omega * ns_t;
         N_VLinearSum(1.0, ns_s.getNVector(), -omega, ns_t.getNVector(), ns_r.getNVector());
 
         // Compute the (unscaled) residual
         N_VProd(ns_r.getNVector(), ns_Jdiag.getNVector(), ns_r.getNVector());
-        res = sqrt(N_VDotProd(ns_r.getNVector(), ns_r.getNVector()));
+        double res = sqrt(N_VDotProd(ns_r.getNVector(), ns_r.getNVector()));
 
         // Test convergence
         if (res < atol) {
